(in-package "HO-VERSION")

(defmacro assert-error (test-form)
  `(assert (handler-case (progn ,test-form nil)
            (error () t))))

(defun test-read-chars ()
  (with-input-from-string (s "")
    (assert (string= "" (read-chars #'alpha-char-p s))))
  (with-input-from-string (s "Hello1")
    (assert (string= "Hello" (read-chars #'alpha-char-p s))))
  (with-input-from-string (s "world ")
    (assert (string= "world" (read-chars #'alpha-char-p s)))))

(defun test-simple-version= ()
  (assert
   (simple-version= (make-simple-version-boa nil 0 0 0)
                    (make-simple-version-boa nil 0 0 0)))
  (assert
   (simple-version= (make-simple-version-boa "V" 0 0 0)
                    (make-simple-version-boa "V" 0 0 0)))
  (assert
   (not
    (simple-version= (make-simple-version-boa nil 0 0 0)
                     (make-simple-version-boa "V" 0 0 0))))
  (assert
   (not
    (simple-version= (make-simple-version-boa "v" 0 0 0)
                     (make-simple-version-boa "V" 0 0 0))))
  (assert
   (not
    (simple-version= (make-simple-version-boa nil 0 0 0)
                     (make-simple-version-boa nil 0 0 1))))
  (assert
   (not
    (simple-version= (make-simple-version-boa nil 0 0 0)
                     (make-simple-version-boa nil 0 1 0))))
  (assert
   (not
    (simple-version= (make-simple-version-boa nil 0 0 0)
                     (make-simple-version-boa nil 1 0 0)))))

(defun test-simple-version>= ()
  (labels ((c (a b d)
             (simple-version>= (apply #'make-simple-version-boa a)
                               (apply #'make-simple-version-boa b)
                               d)))
    (assert (c '(nil 0 0 0) '(nil 0 0 0) :patch))
    (assert (not (c '("x" 0 0 0) '("y" 0 0 0) :patch)))
    (assert (not (c '(nil 1 0 0) '(nil 0 0 0) :patch)))
    (assert (not (c '(nil 0 1 0) '(nil 0 0 0) :patch)))
    (assert (c '(nil 0 0 1) '(nil 0 0 0) :patch))
    
    (assert (c '(nil 0 0 0) '(nil 0 0 0) :minor))
    (assert (c '(nil 0 1 0) '(nil 0 0 0) :minor))
    (assert (c '(nil 0 0 1) '(nil 0 0 0) :minor))
    (assert (not (c '(nil 1 0 0) '(nil 0 0 0) :minor)))
    (assert (c '(nil 1 2 3) '(nil 1 2 3) :minor))
    (assert (c '(nil 1 2 4) '(nil 1 2 3) :minor))
    (assert (not (c '(nil 1 2 3) '(nil 1 2 4) :minor)))
    (assert (c '(nil 1 3 1) '(nil 1 2 3) :minor))
    (assert (not (c '(nil 1 2 3) '(nil 1 3 1) :minor)))
    (assert (not (c '(nil 2 3 4) '(nil 1 2 3) :minor)))
    (assert (not (c '(nil 1 2 3) '(nil 2 3 4) :minor)))
    (assert (not (c '(nil 1 2 3) '("V" 1 2 3) :minor)))
    (assert (not (c '("v" 1  2 3) '("V" 1 2 3) :minor)))
    (assert (c '("v" 1 2 3) '("v" 1 2 3) :minor))

    (assert (c '(nil 0 0 0) '(nil 0 0 0) :major))
    (assert (not (c '("x" 0 0 0) '("y" 0 0 0) :major)))
    (assert (c '(nil 1 0 0) '(nil 0 0 0) :major))
    (assert (c '(nil 0 1 0) '(nil 0 0 0) :major))
    (assert (c '(nil 0 0 1) '(nil 0 0 0) :major))))

(defun test-bump-simple-version ()
  (assert (simple-version=
           (bump-simple-version (make-simple-version-boa nil 1 2 3) :major)
           (make-simple-version-boa nil 2 0 0)))
  (assert (simple-version=
           (bump-simple-version (make-simple-version-boa nil 1 2 3) :minor)
           (make-simple-version-boa nil 1 3 0)))
  (assert (simple-version=
           (bump-simple-version (make-simple-version-boa nil 1 2 3) :patch)
           (make-simple-version-boa nil 1 2 4)))
  (assert (simple-version=
           (bump-simple-version (make-simple-version-boa "V" 1 2 3) :patch)
           (make-simple-version-boa "V" 1 2 4))))

(defun test-change-simple-version-variant ()
  (assert
   (simple-version=
    (change-simple-version-variant (make-simple-version-boa nil 1 2 3)
                                   nil)
    (make-simple-version-boa nil 1 2 3)))
  (assert
   (simple-version=
    (change-simple-version-variant (make-simple-version-boa nil 1 2 3)
                                   "v")
    (make-simple-version-boa "v" 1 2 3)))
  (assert
   (simple-version=
    (change-simple-version-variant (make-simple-version-boa "v" 1 2 3)
                                   nil)
    (make-simple-version-boa nil 1 2 3)))
  (assert
   (simple-version=
    (change-simple-version-variant (make-simple-version-boa "v" 1 2 3)
                                   "x")
    (make-simple-version-boa "x" 1 2 3))))

(defun test-simple-version ()
  (test-simple-version=)
  (test-simple-version>=)
  (test-bump-simple-version)
  (test-change-simple-version-variant))

(defun test-valid-version-part-list? ()
  (assert #v0.0.0)
  (assert #vVAR:0.0.0)
  (assert #v0.0.0/X:1.2.3)
  (assert-error (parse-version ""))
  (assert #vX:0.0.0/Y:1.2.3)
  (assert-error (parse-version 0.0.0/1.2.3)))

(defun test-version= ()
  (assert (version= #v0.0.0 #v0.0.0))
  (assert (not (version= #v0.0.0 #v0.0.1)))
  (assert (version= #v0.0.0/X:1.2.3 #vX:1.2.3)))

(defun test-version>= ()
  (assert (version>= #v0.0.0 #v0.0.0 :minor))
  (assert (version>= #v0.1.1 #v0.0.0 :minor))
  (assert (not (version>= #v1.0.0 #v0.0.0 :minor)))
  (assert (version>= #v0.1.1/X:0.1.1 #v0.0.0 :minor))
  (assert (not (version>= #v1.0.0/X:0.1.1 #v0.0.0 :minor)))
  (assert (not (version>= #v0.1.1/X:1.0.0 #v0.0.0 :minor)))
  (assert (not (version>= #v0.0.0/X:1.2.3/Y:2.3.4
                          #v0.0.0/X:1.1.1/Y:2.0.0 :minor)))
  (assert (not (version>= #v0.0.0/X:1.2.3/Y:3.0.0
                          #v0.0.0/X:1.1.1/Y:2.0.0 :minor)))
  (assert (not (version>= #v0.0.0/X:1.2.3/Z:2.0.0
                          #v0.0.0/X:1.1.1/Y:2.0.0 :minor)))
  (assert (not (version>= #v0.0.0/X:2.0.0/Y:2.0.0
                          #v0.0.0/X:1.1.1/Y:2.0.0 :minor)))
  (assert (not (version>= #v0.0.0/Z:1.2.3/Y:2.0.0
                          #v0.0.0/X:1.1.1/Y:2.0.0 :minor)))
  (assert (version>= #v0.0.0/X:1.1.1/Y:2.0.0/Z:0.1.2
                     #v0.0.0/X:1.1.1/Y:2.0.0 :minor))
  (assert (not (version>= #v0.0.0/X:1.1.1/Y:2.0.0/Z:1.0.0
                          #v0.0.0/X:1.1.1/Y:2.0.0 :minor)))
  (assert (not (version>= #v0.0.0/X:1.1.2/Y:2.0.0/Z:0.0.0
                          #v0.0.0/X:1.1.1/Y:2.0.0 :minor)))
  :; variant
  (assert (version>= #v0.0.0/X:1.1.2/Y:2.0.0/Z:10.0.0
                     #v0.0.0/X:1.1.2/Y:2.0.0/Z:2.0.0
                     :major))
  (assert (not (version>= #v0.0.0/X:1.1.2/Y:2.0.0/Z:10.0.0
                          #v0.0.0/X:1.1.2/Y:2.0.0/z:2.0.0
                          :major)))
  ;; canonical and non-canonical
  (assert (version>= #v0.0.0/X:1.1.2
                     #vX:1.1.2 :minor))
  (assert (version>= #vX:1.1.2
                     #v0.0.0/X:1.1.2 :minor))
  (assert (not (version>= #v0.0.0/X:1.1.2
                          #vY:1.1.2 :minor)))
  (assert-error (version>= #vX:1.1.2
                           #v0.0.0/Y:1.1.2 :minor))
  (assert (version>= #vX:0.1.3
                     #vX:0.1.2 :minor))
  (assert (version>= #v0.0.0/pre:3.2.1/v81:0.3.0 #vpre:3.0.0))
  (assert (not (version>= #v0.0.0/pre:3.2.1/v81:0.3.0 #vv81:0.3.0/Q:3.2.1)))
  (assert (version>= #v0.0.0/pre:3.2.1/v81:0.3.0/Q:3.3.3
                     #vv81:0.3.0/Q:3.2.1))
  (assert-error (version>= #vY:0.1.3
                           #vX:0.1.2 :minor))
  (assert-error (version>= #vX:0.0.0/Y:0.1.3
                           #vZ:0.1.2/A:3.4.5 :minor)))

(defun test-bump-version ()
  (assert (version= (bump-version #v1.2.3 :major)
                    #v2.0.0))
  (assert (version= (bump-version #v1.2.3 :minor)
                    #v1.3.0))
  (assert (version= (bump-version #v1.2.3 :patch)
                    #v1.2.4))
  (assert (version= (bump-version #v4.5.6/X:1.2.3/Y:7.8.9 :major)
                    #v4.5.6/X:1.2.3/Y:8.0.0)))

(defun test-branch-version ()
  (assert (version= (branch-version #v1.2.3 "X")
                    #v1.2.3/X:0.0.0))
  (assert (version= (branch-version #v1.2.3/X:4.5.6 "Y")
                    #v1.2.3/X:4.5.6/Y:0.0.0))
  (assert-error (branch-version #vX:1.2.3 "Y")))

(defun test-version ()
  (test-simple-version)
  (test-valid-version-part-list?)
  (test-version=)
  (test-version>=)
  (test-bump-version)
  (test-branch-version))
